# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, TypeGuard

import numpy as _np

class DPError(Exception):
    pass

integer = int | _np.integer[Any]
floating = float | _np.floating[Any]
realnum = integer | floating

def is_integer(x: Any) -> TypeGuard[integer]:
    return isinstance(x, (int, _np.integer))

def is_floating(x: Any) -> TypeGuard[floating]:
    return isinstance(x, (float, _np.floating))

def is_realnum(x: Any) -> TypeGuard[realnum]:
    return is_integer(x) or is_floating(x)

ElementType = realnum | str | bool
